import abc
import json
from dataclasses import dataclass
from typing import List, Union


class MessageProto(abc.ABC):
    # first item in the array is the version

    def is_version(self, data: List[str]) -> bool:
        if len(data) > 0:
            return data[0] == self.version()
        return False

    def version(self) -> str:
        return self.__class__.__name__

    @abc.abstractmethod
    def _serialize(self) -> List[str]:
        pass

    def serialize(self) -> List[str]:
        return [self.version(), *self._serialize()]

    @classmethod
    @abc.abstractmethod
    def deserialize(cls, data: List[str]) -> "MessageProto":
        pass


@dataclass(kw_only=True)
class RequestMessageV1(MessageProto):
    request_path: str  # 1
    method: str  # 2
    timeout: int  # 3
    payload: Union[str, dict]  # 4

    def __post_init__(self):
        if isinstance(self.payload, dict):
            self.payload = json.dumps(self.payload)

    def _serialize(self) -> List[str]:
        return [self.request_path, self.method, str(self.timeout), self.payload]

    @classmethod
    def deserialize(cls, data: List[str]) -> "RequestMessageV1":
        return cls(
            request_path=data[1], method=data[2], timeout=int(data[3]), payload=data[4]
        )


@dataclass(frozen=True, kw_only=True)
class ResponseMessageV1(MessageProto):
    request_method: str
    request_timeout: int
    response_data: str
    response_status_code: int
    response_content_type: str

    def _serialize(self) -> List[str]:
        return [
            self.request_method,
            str(self.request_timeout),
            self.response_data,
            str(self.response_status_code),
            self.response_content_type,
        ]

    @classmethod
    def deserialize(cls, data: List[str]) -> "ResponseMessageV1":
        return cls(
            request_method=data[1],
            request_timeout=int(data[2]),
            response_data=data[3],
            response_status_code=int(data[4]),
            response_content_type=data[5],
        )



import os
import random
import re
import shutil
import string
from datetime import datetime
from pathlib import Path
from typing import Generator, List, Sequence, Tuple

import pytest
from databricks.sdk import WorkspaceClient
from databricks.sdk.service.catalog import ModelVersionInfo
from mlflow import MlflowClient

import mlflow_extensions.databricks.prebuilt as prebuilt
from mlflow_extensions.databricks.deploy.ez_deploy import (
    EzDeploy,
    EzDeployConfig,
    ServingConfig,
)

ez_deploy_configs: List[EzDeployConfig] = [
    prebuilt.vision.vllm.PHI_3_5_VISION_INSTRUCT_12K_CONFIG,
]


def is_directory_empty(path: Path) -> bool:
    is_empty: bool = not any(path.iterdir())
    return is_empty


@pytest.fixture
def database_name(workspace_client: WorkspaceClient) -> Generator[str, None, None]:
    catalog_name: str = "main"

    dt_string: datetime = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
    random_suffix: str = re.sub(r"[^A-Za-z0-9]", "", dt_string)

    database_name: str = f"mlflow_extensions_test_{random_suffix}"

    if not any([c.name == catalog_name for c in workspace_client.catalogs.list()]):
        workspace_client.catalogs.create(name=catalog_name)

    if not any(
        [
            s.name == database_name
            for s in workspace_client.schemas.list(catalog_name=catalog_name)
        ]
    ):
        workspace_client.schemas.create(catalog_name=catalog_name, name=database_name)

    # Yield to the test
    full_name: str = f"{catalog_name}.{database_name}"
    yield full_name

    workspace_client.schemas.delete(full_name=full_name, force=True)


@pytest.fixture
def config() -> EzDeployConfig:
    config: EzDeployConfig = EzDeployConfig(
        name="test",
        engine_config=None,
        engine_proc=None,
        serving_config=ServingConfig(),
    )
    return config


@pytest.fixture
def deploy(config: EzDeployConfig) -> EzDeploy:
    return EzDeploy(config=config, registered_model_name="catalog.schema.model")


@pytest.mark.unit
def test_should_init(config: EzDeployConfig) -> None:
    _ = EzDeploy(config=config, registered_model_name="catalog.schema.model")


@pytest.mark.unit
def test_init_should_raise_when_registered_model_name_is_not_fully_qualified(
    config: EzDeployConfig,
) -> None:
    with pytest.raises(AssertionError) as ae:
        _ = EzDeploy(config=config, registered_model_name="unqualified_model_name")


@pytest.mark.integration
@pytest.mark.parametrize("config", ez_deploy_configs)
def test_should_download_provided_path(
    config: EzDeployConfig, tmp_path_factory, database_name: str
) -> None:

    registered_model_name: str = f"{database_name}.phi_3_5_vision_instruct_12k"

    deploy: EzDeploy = EzDeploy(
        config=config,
        registered_model_name=registered_model_name,
    )

    dir_name: str = "models"
    download_path: Path = tmp_path_factory.mktemp(dir_name)
    try:
        deploy.download(local_dir=download_path)
        assert deploy._downloaded and not is_directory_empty(download_path)
    finally:
        shutil.rmtree(download_path)


@pytest.mark.unit
def test_should_deploy(deploy: EzDeploy) -> None:
    assert deploy is not None


@pytest.mark.unit
@pytest.mark.parametrize("config", ez_deploy_configs)
def test_register_should_raise_when_not_downloaded(
    deploy: EzDeploy, database_name: str
) -> None:

    registered_model_name: str = f"{database_name}.phi_3_5_vision_instruct_12k"

    deploy: EzDeploy = EzDeploy(
        config=config,
        registered_model_name=registered_model_name,
    )
    with pytest.raises(AssertionError) as ae:
        deploy.register()


@pytest.mark.skipif(
    "DATABRICKS_HOST" not in os.environ,
    reason="Missing environment variable: DATABRICKS_HOST",
)
@pytest.mark.skipif(
    "DATABRICKS_TOKEN" not in os.environ,
    reason="Missing environment variable: DATABRICKS_TOKEN",
)
@pytest.mark.integration
@pytest.mark.databricks
@pytest.mark.parametrize("config", ez_deploy_configs)
def test_register(
    config: EzDeployConfig, workspace_client: WorkspaceClient, database_name: str
) -> None:

    registered_model_name: str = f"{database_name}.phi_3_5_vision_instruct_12k"

    deploy: EzDeploy = EzDeploy(
        config=config,
        registered_model_name=registered_model_name,
    )

    deploy.download()
    deploy.register()

    models: List[ModelVersionInfo] = list(
        workspace_client.model_versions.list(full_name=registered_model_name)
    )
    assert len(models) > 0

    workspace_client.model_versions.delete(
        full_name=registered_model_name, version=models[0].version
    )



import shutil


def get_gpu_count(default: int = 1) -> int:
    try:
        import torch

        return torch.cuda.device_count()
    except ImportError:
        # we are assuming that there is atleast one gpu for when you need to use the engines
        return default


def get_total_shm_size_gb():
    shm_path = "/dev/shm"
    total, used, free = shutil.disk_usage(shm_path)
    return total / (1024**3)


def not_enough_shm(n_gb: int = 1) -> bool:
    try:
        return get_total_shm_size_gb() < n_gb
    except FileNotFoundError:
        return True



# Databricks notebook source
# MAGIC %pip install mlflow-extensions
# MAGIC %pip install mlflow -U
# MAGIC dbutils.library.restartPython()

# COMMAND ----------

from mlflow_extensions.databricks.deploy.ez_deploy import EzDeploy
from mlflow_extensions.databricks.prebuilt import prebuilt

deployer = EzDeploy(
    config=prebuilt.audio.vllm.FIXIE_ULTRA_VOX_0_4_64K_CONFIG,
    registered_model_name="main.default.sri_ultravox_audio_text_to_text_model"
)

deployer.download()

deployer.register()

endpoint_name = "sri_ultravox_audio_text_to_text_model"

deployer.deploy(endpoint_name)

# COMMAND ----------

import requests
import base64


def encode_audio_base64_from_url(audio_url: str) -> str:
    """Encode an audio retrieved from a remote url to base64 format."""

    with requests.get(audio_url) as response:
        response.raise_for_status()
        result = base64.b64encode(response.content).decode('utf-8')

    return result


# gettysburg.wav is a 17 second audio file
audio_data = encode_audio_base64_from_url("https://www2.cs.uic.edu/~i101/SoundFiles/gettysburg.wav")

# COMMAND ----------

from mlflow_extensions.serving.compat.openai import OpenAI
from mlflow_extensions.databricks.prebuilt import prebuilt
from mlflow.utils.databricks_utils import get_databricks_host_creds

workspace_host = spark.conf.get("spark.databricks.workspaceUrl")
endpoint_name = "sri_ultravox_audio_text_to_text_model"
endpoint_url = f"https://{workspace_host}/serving-endpoints/{endpoint_name}/invocations"

token = get_databricks_host_creds().token

client = OpenAI(
    base_url=endpoint_url,
    api_key=token
)

model_name = prebuilt.audio.vllm.FIXIE_ULTRA_VOX_0_4_64K_CONFIG.engine_config.model

chat_completion_from_base64 = client.chat.completions.create(
    messages=[{
        "role":
            "user",
        "content": [
            {
                "type": "text",
                "text": "Breakdown the content of the audio?"
            },
            {
                "type": "audio_url",
                "audio_url": {
                    # Any format supported by librosa is supported
                    "url": f"data:audio/ogg;base64,{audio_data}"
                },
            },
        ],
    }],
    model=model_name,
    max_tokens=512,
)

result = chat_completion_from_base64.choices[0].message.content
print(result)

# COMMAND ----------

from mlflow_extensions.serving.compat.openai import OpenAI
from mlflow_extensions.databricks.prebuilt import prebuilt
from mlflow.utils.databricks_utils import get_databricks_host_creds

workspace_host = spark.conf.get("spark.databricks.workspaceUrl")
endpoint_name = "sri_ultravox_audio_text_to_text_model"
endpoint_url = f"https://{workspace_host}/serving-endpoints/{endpoint_name}/invocations"

token = get_databricks_host_creds().token

client = OpenAI(
    base_url=endpoint_url,
    api_key=token
)

model_name = prebuilt.audio.vllm.FIXIE_ULTRA_VOX_0_4_64K_CONFIG.engine_config.model

from pydantic import BaseModel
from typing import Literal, List


class AudioExtraction(BaseModel):
    year: str
    speaker: str
    location: str
    sentiment: Literal["positive", "negative", "neutral"]
    tone: Literal["somber", "upbeat", "pessemistic"]
    summary: str


chat_completion_from_base64 = client.chat.completions.create(
    messages=[{
        "role":
            "user",
        "content": [
            {
                "type": "text",
                "text": f"Extract the following content from the audio: {str(AudioExtraction.schema())}?"
            },
            {
                "type": "audio_url",
                "audio_url": {
                    # Any format supported by librosa is supported
                    "url": f"data:audio/ogg;base64,{audio_data}"
                },
            },
        ],
    }],
    model=model_name,
    max_tokens=512,
    extra_body={
        "guided_json": AudioExtraction.schema()
    }
)

result = chat_completion_from_base64.choices[0].message.content
print(result)

# COMMAND ----------





import os
import sys
import warnings
from pathlib import Path
from textwrap import dedent

import pytest
from databricks.sdk import WorkspaceClient
from dotenv import find_dotenv, load_dotenv
from mlflow import MlflowClient

src_dir: Path = Path(__file__).parents[1]
sys.path.insert(0, str(src_dir.resolve()))

env_path: str = find_dotenv()
load_dotenv(env_path)


@pytest.fixture
def mlflow_client() -> MlflowClient:
    return MlflowClient()


@pytest.fixture
def workspace_client() -> WorkspaceClient:
    return WorkspaceClient(
        host=os.environ.get("DATABRICKS_HOST"), token=os.environ.get("DATABRICKS_TOKEN")
    )


# Alias HF_TOKEN to HUGGINGFACEHUB_API_TOKEN
if any([e in os.environ for e in ["HUGGINGFACEHUB_API_TOKEN", "HF_TOKEN"]]):
    os.environ["HUGGINGFACEHUB_API_TOKEN"] = os.environ.get(
        "HUGGINGFACEHUB_API_TOKEN", os.environ.get("HF_TOKEN")
    )


def warn_if_missing(*env_vars: str) -> None:
    has_missing: bool = False
    for env_var in env_vars:
        if env_var not in os.environ:
            has_missing = True
            warnings.warn(f"Missing required env var: {env_var}")
    if has_missing:
        warnings.warn(
            dedent(
                """
      Environmnt variables can be set from the shell or a .env file     
    """
            )
        )


warn_if_missing("DATABRICKS_HOST", "DATABRICKS_TOKEN")



import uvicorn
from fastapi import FastAPI, Request, UploadFile
from starlette.responses import StreamingResponse

app = FastAPI()


@app.middleware("http")
async def log_request_data(request: Request, call_next):
    if request.headers.get("Content-Type") == "application/json":
        body = await request.body()
        print(f"Request path: {request.url.path}")
        print(f"Request body: {body.decode('utf-8')}")
    response = await call_next(request)
    return response


@app.api_route("/{full_path:path}", methods=["GET", "POST", "PUT", "DELETE", "PATCH"])
async def handle_any_request(request: Request, full_path: str):
    print(f"Request method: {request.method}")
    print(f"Request URL: {request.url}")
    print(f"Request headers: {request.headers}")
    print(f"Request path: {full_path}")
    if request.headers.get("content-type") == "application/json":
        body = await request.json()
        print(f"Request body: {body}")
    else:
        body = await request.body()
        print(f"Request body: {body.decode('utf-8')}")
    if full_path == "v1/files":
        f = await request.form()
        file: UploadFile = f.get("file")
        if file:

            async def iterfile():
                file_content = await file.read()
                yield file_content

            return StreamingResponse(iterfile(), media_type="text/plain")

    return {}


if __name__ == "__main__":
    uvicorn.run(app, host="127.0.0.1", port=8000)



import json
from dataclasses import asdict, dataclass, field
from typing import List, Literal, Optional, Type

from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import ResourceDoesNotExist
from databricks.sdk.service.serving import EndpointCoreConfigInput, ServedEntityInput

from mlflow_extensions.databricks.deploy.gpu_configs import (
    ALL_VALID_GPUS,
    Cloud,
    GPUConfig,
)
from mlflow_extensions.serving.engines.base import EngineConfig, EngineProcess
from mlflow_extensions.serving.wrapper import (
    ENABLE_DIAGNOSTICS_FLAG,
    CustomServingEnginePyfuncWrapper,
)


@dataclass(kw_only=True, frozen=True)
class ServingConfig:
    valid_gpus: List[GPUConfig] = field(default_factory=lambda: ALL_VALID_GPUS)
    minimum_memory_in_gb: Optional[int] = None

    def smallest_gpu(self, cloud: Cloud) -> GPUConfig:
        filtered_gpus = [gpu for gpu in self.valid_gpus if gpu.cloud == cloud]
        if self.minimum_memory_in_gb is not None:
            filtered_gpus = [
                gpu
                for gpu in filtered_gpus
                if gpu.total_memory_gb > self.minimum_memory_in_gb
            ]
        if len(filtered_gpus) == 0:
            raise ValueError(
                "No valid gpus in cloud please provide the right gpu config"
            )
        return min(filtered_gpus, key=lambda gpu: gpu.total_memory_gb)


@dataclass(frozen=True, kw_only=True)
class EzDeployConfig:
    name: str
    engine_config: EngineConfig
    engine_proc: Type[EngineProcess]
    serving_config: ServingConfig
    pip_config_override: List[str] = None


class EzDeploy:

    def __init__(
        self,
        *,
        config: EzDeployConfig,
        registered_model_name: str,
        databricks_host: str = None,
        databricks_token: str = None,
    ):
        self._config = config
        self._registered_model_name = registered_model_name
        assert len(self._registered_model_name.split(".")) == 3, (
            "Ensure that your registered model name "
            "follows the level namespace;"
            " <catalog>.<schema>.<model_name>"
        )
        self._downloaded = False
        self._model: Optional[CustomServingEnginePyfuncWrapper] = None
        self._latest_registered_model_version = None
        if databricks_host is None or databricks_token is None:
            from mlflow.utils.databricks_utils import get_databricks_host_creds

            self._client = WorkspaceClient(
                host=get_databricks_host_creds().host,
                token=get_databricks_host_creds().token,
            )
            self._cloud = Cloud.from_host(get_databricks_host_creds().host)
        else:
            self._client = WorkspaceClient(host=databricks_host, token=databricks_token)
            self._cloud = Cloud.from_host(databricks_host)

    def download(self, *, local_dir=None):
        self._model = CustomServingEnginePyfuncWrapper(
            engine=self._config.engine_proc, engine_config=self._config.engine_config
        )
        self._model.setup(local_dir=local_dir)
        self._downloaded = True

    def register(self):
        assert self._downloaded is True, "Ensure you run the download method first"
        assert self._model is not None, "Ensure you run download"
        assert (
            self._registered_model_name is not None
        ), "Ensure you provide a valid registered_name"

        import mlflow

        mlflow.set_registry_uri("databricks-uc")

        from mlflow.models import infer_signature

        with mlflow.start_run():
            logged_model = mlflow.pyfunc.log_model(
                "model",
                python_model=self._model,
                artifacts=self._model.artifacts,
                pip_requirements=self._config.pip_config_override
                or self._model.get_pip_reqs(),
                signature=infer_signature(
                    model_input=["string-request"], model_output=["string-response"]
                ),
                registered_model_name=self._registered_model_name,
            )
            mlflow.log_params(
                {
                    "model": self._config.engine_config.model,
                    "host": self._config.engine_config.host,
                    "port": self._config.engine_config.port,
                    "command": str(self._config.engine_config.to_run_command()),
                    "serialized_engine_config": str(
                        json.dumps(asdict(self._config.engine_config))
                    ),
                }
            )
            self._latest_registered_model_version = (
                logged_model.registered_model_version
            )

    def _does_endpoint_exist(self, name) -> bool:
        try:
            self._client.serving_endpoints.get(name)
            return True
        except ResourceDoesNotExist:
            return False

    def deploy(
        self,
        name,
        *,
        scale_to_zero: bool = True,
        workload_size: Literal["Small", "Medium", "Large"] = "Small",
        workload_type: Optional[str] = None,
        enable_diagnostics: bool = False,
    ):
        gpu_cfg = self._config.serving_config.smallest_gpu(self._cloud)
        endpoint_exists = self._does_endpoint_exist(name)
        try:
            from pyspark.sql import SparkSession

            spark = SparkSession.getActiveSession()
            workspace_url = spark.conf.get("spark.databricks.workspaceUrl")
            print(
                f"Deploying model: {name} track progress here: https://{workspace_url}/ml/endpoints/{name}"
            )
        except ImportError:
            print(
                f"Deploying model: {name}; look at the serving tab to track progress..."
            )
        environment_vars = {}
        if enable_diagnostics is True:
            environment_vars[ENABLE_DIAGNOSTICS_FLAG] = "true"
        if endpoint_exists is False:
            self._client.serving_endpoints.create(
                name=name,
                config=EndpointCoreConfigInput(
                    name=name,
                    served_entities=[
                        ServedEntityInput(
                            name=name,
                            entity_name=self._registered_model_name,
                            entity_version=self._latest_registered_model_version,
                            scale_to_zero_enabled=scale_to_zero,
                            workload_type=workload_type or gpu_cfg.name,
                            workload_size=workload_size,
                            environment_vars=environment_vars,
                        )
                    ],
                ),
            )
        else:
            self._client.serving_endpoints.update_config(
                name=name,
                served_entities=[
                    ServedEntityInput(
                        name=name,
                        entity_name=self._registered_model_name,
                        entity_version=self._latest_registered_model_version,
                        scale_to_zero_enabled=scale_to_zero,
                        workload_type=workload_type or gpu_cfg.name,
                        workload_size=workload_size,
                        environment_vars=environment_vars,
                    )
                ],
            )



import warnings

warnings.warn(
    "The module `mlflow_extensions.serving.fixtures` is deprecated. Please use `mlflow_extensions.testing.fixures` instead.",
    DeprecationWarning,
)
from mlflow_extensions.testing.fixures import LocalTestServer  # noqa



import re
from re import Pattern

import pytest

from mlflow_extensions.version import get_mlflow_extensions_version


@pytest.mark.unit
def test_should_return_correct_version() -> None:
    version: str = get_mlflow_extensions_version()
    assert version is not None

    pep440_regex: Pattern = re.compile(
        r"""
    ^
    (?P<version>
        (0|[1-9]\d*)                # Major version
        (\.(0|[1-9]\d*))*           # Minor and Patch versions (optional, can have more)
    )
    (?P<pre>
        ((-|\.)(a|b|rc)             # Pre-release identifier (alpha, beta, rc)
        (0|[1-9]\d*))               # Pre-release number
    )?
    (?P<post>
        (\.post(0|[1-9]\d*))        # Post-release identifier
    )?
    (?P<dev>
        (\.dev(0|[1-9]\d*))         # Development release identifier
    )?
    (?P<local>
        (\+[a-zA-Z0-9]+(\.[a-zA-Z0-9]+)*)   # Local version identifier (optional)
    )?
    $
    """,
        re.VERBOSE,
    )

    match: re.Match = re.match(pep440_regex, version)
    assert version is not None and match is not None
    if match:
        print("Major/Minor:", match.group(1))
        print("Pre-release:", match.group(2))
        print("Post-release:", match.group(3))
        print("Local-Version:", match.group(4))



from importlib.metadata import PackageNotFoundError, version


def get_mlflow_extensions_version():
    try:
        return version("mlflow-extensions")
    except PackageNotFoundError:
        # package is not installed
        return None



from mlflow_extensions.serving.compat import inject_mlflow_openai_compat_client

MODULE = "langchain_openai"
INSTALL = "langchain-openai"

try:
    from langchain_openai import ChatOpenAI, OpenAI
except ImportError as e:
    print(
        f"Error importing {MODULE} module please run "
        f"pip install {INSTALL} or upgrade the sdk by running pip install {INSTALL} --upgrade"
    )

OpenAI = inject_mlflow_openai_compat_client(use_sync=True, use_async=True)(OpenAI)
ChatOpenAI = inject_mlflow_openai_compat_client(use_sync=True, use_async=True)(
    ChatOpenAI
)



from mlflow_extensions.serving.engines.ollama_engine import (
    OllamaEngineConfig,
    OllamaEngineProcess,
)
from mlflow_extensions.serving.engines.sglang_engine import (
    SglangEngineConfig,
    SglangEngineProcess,
)
from mlflow_extensions.serving.engines.vllm_engine import (
    VLLMEngineConfig,
    VLLMEngineProcess,
)



from dataclasses import dataclass
from typing import Dict, List, Union

from mlflow.pyfunc import PythonModelContext

from mlflow_extensions.serving.engines.base import (
    Command,
    EngineConfig,
    EngineProcess,
    debug_msg,
)


@dataclass(frozen=True, kw_only=True)
class MockEngineConfig(EngineConfig):
    def _to_run_command(
        self, context: PythonModelContext = None
    ) -> Union[List[str], Command]:
        return [
            "uvicorn",
            "mlflow_extensions.serving.mock.server:app",
            "--host",
            self.host,
            "--port",
            self.port,
        ]

    def engine_pip_reqs(self, **kwargs) -> List[str]:
        return ["uvicorn", "fastapi"]

    def setup_artifacts(self, local_dir: str = "/root/models") -> Dict[str, str]:
        return {}

    def supported_model_architectures(self) -> List[str]:
        return []


class MockEngineProcess(EngineProcess):
    @property
    def engine_name(self) -> str:
        return "mock-engine"

    def health_check(self) -> bool:
        try:
            resp = self.server_http_client.get("/health")
            return resp.status_code == 200
        except Exception as e:
            debug_msg(
                f"Health check failed with error {e}; server may not be up yet or crashed;"
            )
            return False



from dataclasses import dataclass, field

from mlflow_extensions.databricks.deploy.ez_deploy import EzDeployConfig, ServingConfig
from mlflow_extensions.serving.engines import SglangEngineConfig, SglangEngineProcess

_ENGINE = SglangEngineProcess
_ENGINE_CONFIG = SglangEngineConfig

_llava_model_serving_configs_7b = ServingConfig(
    # rough estimate for the engines this includes model weights + kv cache + overhead + intermediate states
    minimum_memory_in_gb=30,
)

_llava_model_serving_configs_70b = ServingConfig(
    # rough estimate for the engines this includes model weights + kv cache + overhead + intermediate states
    minimum_memory_in_gb=100,
)

# # https://huggingface.co/liuhaotian/llava-v1.5-7b
# LLAVA_1_5_LLAMA_CONFIG = EzDeployConfig(
#     name="llava_1_5_llama_based",
#     engine_proc=_ENGINE,
#     engine_config=_ENGINE_CONFIG(
#         model="liuhaotian/llava-v1.5-7b",
#         tokenizer_path="llava-hf/llava-1.5-7b-hf",
#         chat_template_builtin_name="vicuna_v1.1",
#         verify_chat_template=False,  # it will use the built in chat template
#     ),
#     serving_config=_llava_model_serving_configs_7b,
# )

# # https://huggingface.co/liuhaotian/llava-v1.6-vicuna-7b
# LLAVA_1_6_VICUNA_CONFIG = EzDeployConfig(
#     name="llava_1_6_7b_vicuna_based",
#     engine_proc=_ENGINE,
#     engine_config=_ENGINE_CONFIG(
#         model="liuhaotian/llava-v1.6-vicuna-7b",
#         tokenizer_path="llava-hf/llava-1.5-7b-hf",
#         chat_template_builtin_name="vicuna_v1.1",
#         verify_chat_template=False,  # it will use the built in chat template
#     ),
#     serving_config=_llava_model_serving_configs_7b,
# )

# # https://huggingface.co/liuhaotian/llava-v1.6-34b
# LLAVA_1_6_YI_34B_CONFIG = EzDeployConfig(
#     name="llava_1_6_34b_yi_based",
#     engine_proc=_ENGINE,
#     engine_config=_ENGINE_CONFIG(
#         model="liuhaotian/llava-v1.6-34b",
#         tokenizer_path="liuhaotian/llava-v1.6-34b-tokenizer",
#     ),
#     serving_config=_llava_model_serving_configs_7b,
# )

# https://huggingface.co/lmms-lab/llama3-llava-next-8b
LLAVA_NEXT_LLAMA3_8B_CONFIG = EzDeployConfig(
    name="llava_next_llama3_8b_based",
    engine_proc=_ENGINE,
    engine_config=_ENGINE_CONFIG(
        model="lmms-lab/llama3-llava-next-8b",
        chat_template_builtin_name="llava_llama_3",
    ),
    serving_config=_llava_model_serving_configs_7b,
)


# https://huggingface.co/lmms-lab/llava-next-72b
LLAVA_NEXT_QWEN_1_5_72B_CONFIG = EzDeployConfig(
    name="llava_next_qwen_1_5_72b_based",
    engine_proc=_ENGINE,
    engine_config=_ENGINE_CONFIG(
        model="lmms-lab/llama3-llava-next-8b",
        chat_template_builtin_name="chatml-llava",
    ),
    serving_config=_llava_model_serving_configs_70b,
)

# https://huggingface.co/lmms-lab/llava-onevision-qwen2-7b-ov
LLAVA_ONEVISION_QWEN_2_7B_CONFIG = EzDeployConfig(
    name="llava_onevision_qwen_2_7b_based",
    engine_proc=_ENGINE,
    engine_config=_ENGINE_CONFIG(
        model="lmms-lab/llava-onevision-qwen2-7b-ov",
        chat_template_builtin_name="chatml-llava",
    ),
    serving_config=_llava_model_serving_configs_7b,
)


# https://huggingface.co/lmms-lab/llava-onevision-qwen2-72b-ov
LLAVA_ONEVISION_QWEN_2_72B_CONFIG = EzDeployConfig(
    name="llava_onevision_qwen_2_72b_based",
    engine_proc=_ENGINE,
    engine_config=_ENGINE_CONFIG(
        model="lmms-lab/llava-onevision-qwen2-72b-ov",
        chat_template_builtin_name="chatml-llava",
    ),
    serving_config=_llava_model_serving_configs_70b,
)


@dataclass(frozen=True)
class SglangVision:
    LLAVA_NEXT_LLAMA3_8B: EzDeployConfig = field(
        default_factory=lambda: LLAVA_NEXT_LLAMA3_8B_CONFIG
    )

    LLAVA_NEXT_QWEN_1_5_72B_CONFIG: EzDeployConfig = field(
        default_factory=lambda: LLAVA_NEXT_QWEN_1_5_72B_CONFIG
    )

    LLAVA_ONEVISION_QWEN_2_7B_CONFIG: EzDeployConfig = field(
        default_factory=lambda: LLAVA_ONEVISION_QWEN_2_7B_CONFIG
    )

    LLAVA_ONEVISION_QWEN_2_72B_CONFIG: EzDeployConfig = field(
        default_factory=lambda: LLAVA_ONEVISION_QWEN_2_72B_CONFIG
    )



from dataclasses import dataclass, field

from mlflow_extensions.databricks.deploy.ez_deploy import EzDeployConfig, ServingConfig
from mlflow_extensions.serving.engines import VLLMEngineConfig, VLLMEngineProcess

_ENGINE = VLLMEngineProcess
_ENGINE_CONFIG = VLLMEngineConfig

# https://huggingface.co/fixie-ai/ultravox-v0_4
FIXIE_ULTRA_VOX_0_4_64K_CONFIG = EzDeployConfig(
    name="fixie_utravox_0_4_llama_3_1_whisper_medium",
    engine_proc=_ENGINE,
    engine_config=_ENGINE_CONFIG(
        model="fixie-ai/ultravox-v0_4",
        trust_remote_code=True,
        guided_decoding_backend="outlines",
        max_model_len=64000,
    ),
    serving_config=ServingConfig(
        # rough estimate for the engines this includes model weights + kv cache + overhead + intermediate states
        minimum_memory_in_gb=60,
    ),
)

# https://huggingface.co/fixie-ai/ultravox-v0_4
FIXIE_ULTRA_VOX_0_4_128K_CONFIG = EzDeployConfig(
    name="fixie_utravox_0_4_llama_3_1_whisper_medium",
    engine_proc=_ENGINE,
    engine_config=_ENGINE_CONFIG(
        model="fixie-ai/ultravox-v0_4",
        trust_remote_code=True,
        guided_decoding_backend="outlines",
    ),
    serving_config=ServingConfig(
        # rough estimate for the engines this includes model weights + kv cache + overhead + intermediate states
        minimum_memory_in_gb=110,
    ),
)


@dataclass(frozen=True)
class VllmAudio:
    FIXIE_ULTRA_VOX_0_4_64K_CONFIG: EzDeployConfig = field(
        default_factory=lambda: FIXIE_ULTRA_VOX_0_4_64K_CONFIG
    )
    FIXIE_ULTRA_VOX_0_4_128K_CONFIG: EzDeployConfig = field(
        default_factory=lambda: FIXIE_ULTRA_VOX_0_4_128K_CONFIG
    )



from dataclasses import dataclass, field

from mlflow_extensions.databricks.deploy.ez_deploy import EzDeployConfig, ServingConfig
from mlflow_extensions.serving.engines import SglangEngineConfig, SglangEngineProcess

_ENGINE = SglangEngineProcess
_ENGINE_CONFIG = SglangEngineConfig

# Requires HF token
# https://huggingface.co/google/gemma-2-9b-it
GEMMA_2_9B_IT_CONFIG = EzDeployConfig(
    name=f"gemma_2_9b_it",
    engine_proc=_ENGINE,
    engine_config=_ENGINE_CONFIG(model="google/gemma-2-9b-it"),
    serving_config=ServingConfig(
        # rough estimate for the engines this includes model weights + kv cache + overhead + intermediate states
        minimum_memory_in_gb=30,
    ),
)

# Requires HF token
# https://huggingface.co/meta-llama/Meta-Llama-3.1-8B
META_LLAMA_3_1_8B_INSTRUCT_CONFIG = EzDeployConfig(
    name=f"meta_llama_3_1_8b_instruct",
    engine_proc=_ENGINE,
    engine_config=_ENGINE_CONFIG(model="meta-llama/Meta-Llama-3.1-8B-Instruct"),
    serving_config=ServingConfig(
        # rough estimate for the engines this includes model weights + kv cache + overhead + intermediate states
        minimum_memory_in_gb=30,
    ),
)


@dataclass(frozen=True)
class SglangText:
    GEMMA_2_9B_IT: EzDeployConfig = field(default_factory=lambda: GEMMA_2_9B_IT_CONFIG)
    META_LLAMA_3_1_8B_INSTRUCT_CONFIG: EzDeployConfig = field(
        default_factory=lambda: META_LLAMA_3_1_8B_INSTRUCT_CONFIG
    )



from mlflow_extensions.databricks.deploy.ez_deploy import EzDeployConfig


def make_process_and_get_artifacts(config: EzDeployConfig, local_dir=None):
    if local_dir is not None:
        artifacts = config.engine_config.setup_artifacts(local_dir)
    else:
        artifacts = config.engine_config.setup_artifacts()

    engine = config.engine_proc(config=config.engine_config)

    return engine, artifacts



# Databricks notebook source
# MAGIC %pip install mlflow-extensions
# MAGIC %pip install mlflow -U
# MAGIC dbutils.library.restartPython()

# COMMAND ----------

from mlflow_extensions.databricks.deploy.ez_deploy import EzDeploy
from mlflow_extensions.databricks.prebuilt import prebuilt

deployer = EzDeploy(
  config=prebuilt.text.vllm.NUEXTRACT,
  registered_model_name="main.default.nuextract_vllm"
)

deployer.download()

deployer.register()

endpoint_name = "nuextract_vllm"

deployer.deploy(endpoint_name)


# COMMAND ----------

from mlflow_extensions.serving.compat.openai import OpenAI
from mlflow_extensions.databricks.prebuilt import prebuilt
from mlflow.utils.databricks_utils import get_databricks_host_creds

workspace_host = spark.conf.get("spark.databricks.workspaceUrl")
endpoint_name = "nuextract_vllm"
endpoint_url = f"https://{workspace_host}/serving-endpoints/{endpoint_name}/invocations"

token = get_databricks_host_creds().token

client = OpenAI(
  base_url=endpoint_url,
  api_key=token
)

model_name = prebuilt.text.vllm.NUEXTRACT.engine_config.model

from pydantic import BaseModel
from typing import Literal, List

class ExtractedBody(BaseModel):
    product: str
    languages: Literal["python", "sql", "scala"]
    keywords: List[str]
    strategies: List[str]


response = client.chat.completions.create(
    model=model_name,
    messages=[
        {
            "role": "user",
            "content": """The Databricks Lakehouse Platform for Dummies is your guide to simplifying 
your data storage. The lakehouse platform has SQL and performance 
capabilities - indexing, caching and MPP processing - to make 
BI work rapidly on data lakes. It also provides direct file access 
and direct native support for Python, data science and 
AI frameworks without the need to force data through an 
SQL-based data warehouse. Find out how the lakehouse platform 
creates an opportunity for you to accelerate your data strategy."""
        }
    ],
    extra_body={
        "guided_json": ExtractedBody.schema()
    }
)
response.choices[0].message.content

# COMMAND ----------


